#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#define GL_GLEXT_PROTOTYPES	1
#include <GL/glut.h>
#include <time.h>
#include <sys/time.h>
#include <assert.h>


void redraw(void);
void draw_scene(unsigned long msec);
void fract_sph(float sz, int dir, int iter);
void idle_handler(void);
void key_handler(unsigned char key, int x, int y);
unsigned long get_msec(void);
unsigned int load_tex(const char *fname);
int next_pow(int x);

int wire = 0;
int sep = 0;
int blur = 1;
int max_iter = 4;
float scale = 0.4;
unsigned int env_map, rtarg;

int scrx = 800;
int scry = 600;
int rtargx, rtargy;


int main(int argc, char **argv) {	
	int i;
	
	glutInitWindowSize(scrx, scry);
	glutInit(&argc, argv);

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'b':
				blur = !blur;
				break;

			case 'i':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-i must be followed by a number\n");
					return EXIT_FAILURE;
				}
				max_iter = atoi(argv[i]);
				break;

			case 's':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-s must be followed by a number\n");
					return EXIT_FAILURE;
				}
				scale = atof(argv[i]);
				break;
			}
		}
	}
	
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow("foo");

	scrx = glutGet(GLUT_WINDOW_WIDTH);
	scry = glutGet(GLUT_WINDOW_HEIGHT);

	rtargx = next_pow(scrx);
	rtargy = next_pow(scry);

	glutDisplayFunc(redraw);
	glutIdleFunc(idle_handler);
	glutKeyboardFunc(key_handler);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0, 1.33333, 1.0, 1000.0);
	glMatrixMode(GL_MODELVIEW);

	glClearColor(0, 0, 0, 1);
	glClearDepth(1.0);

	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);

	if(!(env_map = load_tex("refmap2.ppm"))) {
		return EXIT_FAILURE;
	}

	/* setup convolution kernels */
	{
#define KSZ		5
#define SVAL 	(1.0 / (KSZ * KSZ))
#define SSVAL 	(1.0 / KSZ)

		int i;
		float kernel[KSZ * KSZ * 3];
		float sep_kern[KSZ * 3];
		float scale_vec[4] = {SVAL, SVAL, SVAL, SVAL};
		float sep_scale[4] = {SSVAL, SSVAL, SSVAL, SSVAL};
		int max_conv_x, max_conv_y;

		for(i=0; i<KSZ * KSZ * 3; i++) {
			kernel[i] = 1.0;
			if(i < KSZ * 3) {
				sep_kern[i] = 1.0;
			}
		}

		glGetConvolutionParameteriv(GL_CONVOLUTION_2D, GL_MAX_CONVOLUTION_WIDTH, &max_conv_x);
		glGetConvolutionParameteriv(GL_CONVOLUTION_2D, GL_MAX_CONVOLUTION_HEIGHT, &max_conv_y);

		printf("Max convolution kernel size: %dx%d\n", max_conv_x, max_conv_y);

		glConvolutionParameterfv(GL_CONVOLUTION_2D, GL_CONVOLUTION_FILTER_SCALE, scale_vec);
		glConvolutionParameterfv(GL_SEPARABLE_2D, GL_CONVOLUTION_FILTER_SCALE, sep_scale);
	
		glConvolutionFilter2D(GL_CONVOLUTION_2D, GL_RGB, KSZ, KSZ, GL_RGB, GL_FLOAT, kernel);
		glSeparableFilter2D(GL_SEPARABLE_2D, GL_RGB, KSZ, KSZ, GL_RGB, GL_FLOAT, sep_kern, sep_kern);
	}

	/* create render target texture */
	glGenTextures(1, &rtarg);
	glBindTexture(GL_TEXTURE_2D, rtarg);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, rtargx, rtargy, 0, GL_BGRA, GL_UNSIGNED_BYTE, 0);

	glutMainLoop();
	return 0;
}

void redraw(void) {
	unsigned long msec = get_msec();
	float umax = (float)scrx / (float)rtargx;
	float vmax = (float)scry / (float)rtargy;
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	draw_scene(msec);

	if(blur) {
		glEnable(sep ? GL_SEPARABLE_2D : GL_CONVOLUTION_2D);

		glBindTexture(GL_TEXTURE_2D, rtarg);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, scrx, scry);
		glDisable(sep ? GL_SEPARABLE_2D : GL_CONVOLUTION_2D);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_LIGHTING);	
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_TEXTURE_2D);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex2f(-1, -1);
		glTexCoord2f(umax, 0);
		glVertex2f(1, -1);
		glTexCoord2f(umax, vmax);
		glVertex2f(1, 1);
		glTexCoord2f(0, vmax);
		glVertex2f(-1, 1);
		glEnd();
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);	

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}

	glutSwapBuffers();
}

void draw_scene(unsigned long msec) {
	static int dlist = 0;
	float dcol[] = {0.4, 0.4, 0.4, 1.0};
	float scol[] = {1.0, 1.0, 1.0, 1.0};
	
	glLoadIdentity();
	glTranslatef(0, 0, -6);
	glRotatef(30, 1, 0, 0);
	glRotatef((float)msec / 100.0, 0, 1, 0);

	if(!dlist) {
		fprintf(stderr, "Generating geometry\n");
		dlist = glGenLists(2);
		glNewList(dlist, GL_COMPILE);
		fract_sph(1.0, 0, max_iter);
		glEndList();

		wire = 1;
		glNewList(dlist + wire, GL_COMPILE);
		fract_sph(1.0, 0, max_iter);
		glEndList();
		wire = 0;
	}

	/* first pass, diffuse/specular */
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, dcol);
	glMaterialfv(GL_FRONT, GL_SPECULAR, scol);
	glMaterialf(GL_FRONT, GL_SHININESS, 60.0);
	glCallList(dlist + wire);

	/* second pass, envmap */
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	glDepthFunc(GL_LEQUAL);
	
	glBindTexture(GL_TEXTURE_2D, env_map);
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glCallList(dlist + wire);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_2D);

	glDepthFunc(GL_LESS);
	glDisable(GL_BLEND);
}

void fract_sph(float sz, int dir, int iter) {
	enum {CENTER, PX, NX, PY, NY, PZ, NZ};
	float offs = sz + sz * scale;
	int detail;
	
	if(!iter) return;
	
	detail = 20 - (max_iter - iter) * 4;
	if(detail < 8) detail = 8;

	if(wire) {
		glutWireSphere(sz, detail, detail / 2);
	} else {
		glutSolidSphere(sz, detail, detail / 2);
	}
	
	if(dir != NX) {
		glPushMatrix();
		glTranslatef(offs, 0, 0);
		fract_sph(sz * scale, PX, iter - 1);	
		glPopMatrix();
	}

	if(dir != PX) {
		glPushMatrix();
		glTranslatef(-offs, 0, 0);
		fract_sph(sz * scale, NX, iter - 1);
		glPopMatrix();
	}

	if(dir != NY) {
		glPushMatrix();
		glTranslatef(0, offs, 0);
		fract_sph(sz * scale, PY, iter - 1);
		glPopMatrix();
	}

	if(dir != PY) {
		glPushMatrix();
		glTranslatef(0, -offs, 0);
		fract_sph(sz * scale, NY, iter - 1);
		glPopMatrix();
	}

	if(dir != NZ) {
		glPushMatrix();
		glTranslatef(0, 0, offs);
		fract_sph(sz * scale, PZ, iter - 1);
		glPopMatrix();
	}

	if(dir != PZ) {
		glPushMatrix();
		glTranslatef(0, 0, -offs);
		fract_sph(sz * scale, NZ, iter - 1);
		glPopMatrix();
	}
}


void idle_handler(void) {
	glutPostRedisplay();
}

void key_handler(unsigned char key, int x, int y) {
	switch(key) {
	case 27:
	case 'q':
	case 'Q':
		exit(0);

	case 'w':
	case 'W':
		wire = !wire;
		break;

	case 's':
		sep = !sep;
		printf("convolution mode: %s\n", (sep ? "separable" : "2D"));
		break;
	}
}

unsigned long get_msec(void) {
	static struct timeval timeval, first_timeval;
	
	gettimeofday(&timeval, 0);
	if(first_timeval.tv_sec == 0) {
		first_timeval = timeval;
		return 0;
	}
	return (timeval.tv_sec - first_timeval.tv_sec) * 1000 + (timeval.tv_usec - first_timeval.tv_usec) / 1000;
}

int check_ppm(FILE *fp);
void *load_ppm(FILE *fp, unsigned long *xsz, unsigned long *ysz);

unsigned int load_tex(const char *fname) {
	FILE *fp;
	void *img;
	unsigned long xsz, ysz;
	unsigned int tex;

	if(!(fp = fopen(fname, "rb")) || !check_ppm(fp)) {
		return 0;
	}

	if(!(img = load_ppm(fp, &xsz, &ysz))) {
		return 0;
	}

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, xsz, ysz, 0, GL_BGRA, GL_UNSIGNED_BYTE, img);

	free(img);
	return tex;
}

int next_pow(int x) {
	int test = 1;
	while(test < x) {
		test <<= 1;
	}
	return test;
}
